import time
import warnings
import geopandas as gpd
import matplotlib.pyplot as plt
import numpy as np
import overpy
from scipy.spatial.distance import cdist
from shapely.geometry import LineString, Point, Polygon, box
from shapely.ops import unary_union
import os
import shutil
import urllib.request
import zipfile
from typing import List
import pandas as pd
import tempfile

def make_overpy_request(request_body: str, retries: int):
    for _ in range(retries):
        try:
            api = overpy.Overpass()
            result = api.query(request_body)
        except (
            overpy.exception.OverpassTooManyRequests,
            overpy.exception.OverpassGatewayTimeout,
        ):
            warnings.warn("Overpass Too Many Requests --- Retry in 5 seconds")
            time.sleep(5)
        else:
            return result


def get_geoseries_streets(
    measurement_coords: gpd.GeoSeries, retries: int = 5, crop: bool = False
) -> gpd.GeoSeries:
    """
    Obtain the street shapes in the area spanned by the measurement_coords.

    :param measurement_coords: A geopandas geoseries of Points(lon, lat) representing the measurement positions in a EPSG:4326 projection
    :param retries: Number of retries for overpy requests, default:5
    :param crop: Set to True if the returned geoseries should be fixed to bounding box determined by measurement_coords, default: False

    :return: A geopandas geoseries consiting of LineStrings representing the streets in EPSG:4326
    """

    if measurement_coords.crs != "EPSG:4326":
        raise ValueError("Make sure to pass data with EPSG:4326 projection")

    # Find total bounds: (lat_min, long_min, lat_max, long_max)
    bounds = tuple(measurement_coords.total_bounds[[1, 0, 3, 2]])

    result = make_overpy_request(
        f"""
        [out:json][timeout:25];
        (
        way['highway']['highway'!='footway']['highway'!='platform']['highway'!='steps']{bounds};
        );
        out body;
        >;
        out skel qt;
        """,
        retries,
    )

    street_series = gpd.GeoSeries(
        LineString(((node.lon, node.lat) for node in way.nodes if len(way.nodes) > 1))
        for way in result.ways
    ).set_crs("EPSG:4326")

    if crop == True:
        street_series = crop_geoseries_by_bounds(street_series, measurement_coords)

    street_series = street_series.explode() # if any MultiLineStrings, they are transformed to LineStrings
    return street_series


def get_geoseries_blockages(
    measurement_coords: gpd.GeoSeries, retries: int = 5
) -> gpd.GeoSeries:
    """
    Obtain the blockage shapes in the area spanned by the measurement_coords.

    :param measurement_coords: A geopandas geoseries of Points(lon, lat) representing the measurement positions in a EPSG:4326 projection

    :return: A geopandas geoseries consiting of Polygons representing the blockages in EPSG:4326
    """

    if measurement_coords.crs != "EPSG:4326":
        raise ValueError("Make sure to pass data with EPSG:4326 projection")

    # Find total bounds: (lat_min, long_min, lat_max, long_max)
    bounds = tuple(measurement_coords.total_bounds[[1, 0, 3, 2]])

    result = make_overpy_request(
        f"""
        [out:json][timeout:25];
        (
        way['building']{bounds};
        relation["building"]{bounds};
        );
        out body;
        >;
        out skel qt;
        """,
        retries,
    )

    blockages = gpd.GeoSeries(
        Polygon(
            (
                (node.lon, node.lat) for node in way.nodes if len(way.nodes) > 2
            )  # Polygon needs at least 3 nodes
        )
        for way in result.ways
    ).set_crs("EPSG:4326")

    return blockages


def project_onto_streets(
    point_series: gpd.GeoSeries,
    street_series: gpd.GeoSeries,
    epsg: str = "EPSG:31287",
    plot: bool = False,
) -> (gpd.GeoSeries, gpd.GeoSeries):
    """
    Projects each point from the point_series onto the closest element of
    street_series according to a normal projection. For that, both
    series will be transformed into a 2D projection (default EPSG:31287).

    :param point_series: GeoSeries of all points to project onto streets.
    :param street_series: GeoSeries of linestrings representing the streets.
    :param epsg: EPSG projection to use during the projection operation.
    :param plot: Boolean to indiceate whether the result of the operation shall be plotted.

    :return: A tuple of two GeoSeries containing the result of the operation as well as the deviation introduced.
    """

    if epsg == "EPSG:4326":
        raise ValueError(
            "Projection cannot be conducted using geographical coordinates such as EPSG:4326. Select 2D projection instead."
        )

    if point_series.crs is None or street_series.crs is None:
        raise ValueError(
            "CRS needs to be set explicitly for passed Geoseries: Use .set_crs()."
        )

    point_original_crs, street_original_crs = point_series.crs, street_series.crs
    point_series = point_series.to_crs(epsg)
    street_series = street_series.to_crs(epsg)

    projected = []
    for point in point_series:
        street_ind = street_series.distance(point).argmin()
        projected.append(
            street_series.loc[street_ind].interpolate(
                street_series.loc[street_ind].project(point)
            )
        )

    projected = gpd.GeoSeries(projected)
    projected = projected.set_crs(epsg)
    deviation_projection = projected.distance(point_series)

    if plot:
        plt.figure()
        plt.title("Deviation Histogram")
        deviation_projection.plot.hist(bins=50)
        plt.show()

        fig, (ax1, ax2) = plt.subplots(1, 2, figsize=(20, 10))
        ax1.set_title("Raw Measurements")
        ax2.set_title("After Projection")
        ax1.scatter(point_series.x, point_series.y, label="Point Series")
        clb = ax2.scatter(
            projected.x, projected.y, c=deviation_projection, label="Projected Series"
        )
        for street in street_series:
            ax1.plot(*street.xy, color="black")
            ax2.plot(*street.xy, color="black")
        bar = fig.colorbar(clb)
        bar.ax.set_title("Deviation [Meters]")
        ax1.legend()
        ax2.legend()
        plt.show()

    #Reset Street and Point Series CRS
    street_series.to_crs(street_original_crs)
    point_series.to_crs(point_original_crs)
    projected.to_crs(point_original_crs)

    return projected, deviation_projection


def crop_geoseries_by_bounds(
    original_geoseries: gpd.GeoSeries, filter_geoseries: gpd.GeoSeries
) -> gpd.GeoSeries:
    """
    Returns cut off geoseries from 'original_geoseries' which lies within the bounds on the filter_geoseries

    :param original_geoseries: A geopandas geoseries of points/streets/polygons to be filtered, represented in a EPSG:4326 projection
    :param filter_geoseries: A geopandas geoseries based on which a bounding box is for filter is made

    :return: geopandas geoseries of same format as original_geoseries
    """
    if original_geoseries.crs != "EPSG:4326" or filter_geoseries.crs != "EPSG:4326":
        raise ValueError("Make sure to pass data with EPSG:4326 projection")

    filter_bounds = filter_geoseries.total_bounds
    bound_box = box(
        filter_bounds[0], filter_bounds[1], filter_bounds[2], filter_bounds[3]
    )
    bound_box = gpd.GeoDataFrame(geometry=gpd.GeoSeries(bound_box)).set_crs(
        "EPSG:4326"
    )  # create geodataframe for overlay
    filtered_series = (
        gpd.overlay(
            gpd.GeoDataFrame(geometry=original_geoseries), bound_box, how="intersection"
        )
    ).geometry

    return filtered_series


def interpolate_series_at_distance(
    street_series: gpd.GeoSeries, distance: int = 2
) -> gpd.GeoSeries:
    """
    Interpolates along the geoseries streets_series to get nodes at a distance 'distance' apart.
    Make sure to pass data in Cartesian coordinates (e.g. use street_series.to_crs('EPSG:31287') for Austria).

    :param street_series: geopandas geoseries
    :param distance: integer determining the spacing between interpolation points, default=2

    :return: geopandas geoseries overlapping with original geoseries containing more frequent nodes
    """
    if street_series.crs == "EPSG:4326":
        raise ValueError(
            "Make sure to pass data in Cartesian coordinates (e.g. use street_series.to_crs('EPSG:31287') for Austria)"
        )

    if distance <= 0:
        raise ValueError("Distance has to be a positive number.")

    lines = []
    for linestring in street_series:
        list_points = []  # list to hold all the point coords
        current_dist = distance  # set the current distance to place the point
        line_length = (
            linestring.length
        )  # get the total length of the line (not number of points, but actual length)
        list_points.append(
            Point(list(linestring.coords[0]))
        )  # append the starting coordinate to the list
        while (
            current_dist < line_length
        ):  # while the current cumulative distance is less than the total length of the line
            list_points.append(
                linestring.interpolate(current_dist)
            )  # use interpolate and increase the current distance
            current_dist += distance
        list_points.append(
            Point(list(linestring.coords[-1]))
        )  # append end coordinate to the list
        line = LineString(list_points)
        lines.append(line)

    interpolated_street_series = gpd.GeoSeries(lines)
    return interpolated_street_series


def dissolve_touching_polygons(polygon_series: gpd.GeoSeries) -> gpd.GeoSeries:
    """
    Returns geoseries of dissolved touching polygons with closed polygon holes.

    :param polygon_series: geopandas geoseries of polygons

    :return: geopandas geoseries of dissolved polygons
    """
    poly_list = [
        polygon_series[i]
        for i in range(polygon_series.size)
        if polygon_series[i] != None
    ]
    poly_union = unary_union(
        poly_list
    )  # MultiPolygon containing polygon object sequence
    p_list = []
    for geom in poly_union.geoms:  # loop over polygon object sequence
        p = Polygon(
            geom.exterior
        )  # create new polygon only using the exterior (closes holes)
        p_list.append(p)
    polygon_series_disolved = gpd.GeoSeries(p_list)

    return polygon_series_disolved


def delete_indoor_meas(
    measurement_coords: gpd.GeoSeries, polygons: gpd.GeoSeries
) -> gpd.GeoSeries:
    """
    Removes indoor measurements by deleting all measurement points inside the polygons.

    :param measurement_coords: A geopandas geoseries of Points(lon, lat) representing the measurement positions in a EPSG:4326 projection
    :param polygons: A geopandas geoseries of Polygons in EPSG:4326 projection

    :return: A geopandas geoseries of Points(lon, lat) representing the outdoor measurement positions in a EPSG:4326 projection

    """
    meas_gdf = gpd.GeoDataFrame(geometry=measurement_coords)
    polygons_gdf = gpd.GeoDataFrame(geometry=polygons)

    indoor_meas_gdf = gpd.sjoin(meas_gdf, polygons_gdf)
    outdoor_meas_gdf = meas_gdf.drop(indoor_meas_gdf.index).reset_index()

    return outdoor_meas_gdf.geometry


def pull_from_geodatenviewer_list(squares: List[str]) -> gpd.GeoDataFrame:
    """
    Downloads .zip files for specified squares from https://www.wien.gv.at/ma41datenviewer/public/,
    extracts and combines all .shp files to single geopandas geodataframe, then deletes the dir_path directory.

    :param dir_path: path to the directory where the files will be downloaded and extracted to
    :param squares: list of integers denoting the squares to be downloaded

    :return: geopandas geodataframe containing all data from .shp files
    """

    dir_path = tempfile.mkdtemp() # create temporary directory

    # downloading .zip files
    # example of single square with data: https://www.wien.gv.at/ma41datenviewer/downloads/ma41/geodaten/fmzk_bkm/103081_bkm.zip
    for square_number in squares:
        urllib.request.urlretrieve(
            "https://www.wien.gv.at/ma41datenviewer/downloads/ma41/geodaten/fmzk_bkm/{}_bkm.zip".format(
                square_number
            ),
            os.path.join(dir_path, "{}.zip".format(square_number)),
        )

    # extracting and deleting .zip files
    for item in os.listdir(dir_path):  # loop through items in dir
        if item.endswith(".zip"):
            file_name = os.path.join(dir_path, item)
            zip_ref = zipfile.ZipFile(file_name)  # create zipfile object
            zip_ref.extractall(dir_path)  # extract file to dir
            zip_ref.close()
            os.remove(file_name)  # delete zipped file

    # combine all .shp files
    geo_df_all = pd.DataFrame()
    for square in squares:
        geo_df = gpd.read_file(os.path.join(dir_path, square + "_bkm.shp"))
        geo_df_all = pd.concat([geo_df_all, geo_df], ignore_index=True)

    shutil.rmtree(dir_path)  # deletes the directory containing all the files
    return geo_df_all


def pull_from_geodatenviewer_meas(measurement_coords: gpd.GeoSeries) -> gpd.GeoDataFrame:
    '''
    Downloads raster .zip file for Vienna Austria, extracts all raster polygon IDs containing measurements,
    then calls pull_from_geodatenviewer_list() to extract building polygons.

    :param measurement_coords: geopandas geoseries containing measurements in EPSG:4326 projection

    :return: geopandas geodataframe containing all data from .shp files
    '''

    if measurement_coords.crs != "EPSG:4326":
        raise ValueError("Make sure to pass data with EPSG:4326 projection")

    raster_path = tempfile.mkdtemp()
    urllib.request.urlretrieve(
            "https://data.wien.gv.at/daten/geo?service=WFS&request=GetFeature&version=1.1.0&typeName=ogdwien:MZKBLATT1000OGD&srsName=EPSG:4326&outputFormat=shape-zip",
            os.path.join(raster_path, "raster.zip"),
        )

    file_name = os.path.join(raster_path, "raster.zip")
    zip_ref = zipfile.ZipFile(file_name)  # create zipfile object
    zip_ref.extractall(raster_path)  # extract file to dir
    zip_ref.close()
    os.remove(file_name)  # delete zipped file

    for file in os.listdir(raster_path):
        if file.endswith(".shp"):
            	raster_gdf = gpd.read_file(os.path.join(raster_path, file))
    if raster_gdf.geometry.crs != 'EPSG:4326':
        raise ValueError('Raster EPSG has changed, see https://www.data.gv.at/katalog/dataset/b2d17060-b2f4-4cd7-a2e5-64beccfeb4c1 for mor information.')

    meas_gdf = gpd.GeoDataFrame(geometry=measurement_coords)

    poly_gdf = gpd.sjoin(raster_gdf, meas_gdf, op = 'contains') # find all polygons from raster that contain measurements

    polygonID_list = poly_gdf.MZK1000.unique()
    return_gdf = pull_from_geodatenviewer_list(polygonID_list)

    return return_gdf
